@@ -20,8 +20,8 @@ except ImportError:
     from ordereddict import OrderedDict  # noqa
 
 import collections
-import weakref
 
+from debtcollector import removals
 import prettytable
 import six
 
@@ -73,29 +73,37 @@ class FiniteMachine(object):
         return cls._Effect(new_state['reactions'].get(event),
                            new_state["terminal"])
 
-    @staticmethod
-    def _generate_runner(instance):
-        return _FiniteRunner(instance)
-
-    def __init__(self, start_state):
+    @removals.removed_kwarg('default_start_state',
+                            message="The usage of 'default_start_state' via"
+                            " the machine constructor is deprecated and will"
+                            " be removed in a future version; usage of"
+                            " the 'default_start_state' property setter is"
+                            " recommended.")
+    def __init__(self, default_start_state=None):
         self._transitions = {}
         self._states = OrderedDict()
-        self._start_state = start_state
+        self._default_start_state = default_start_state
         self._current = None
-        self._frozen = False
-        self._runner = self._generate_runner(self)
+        self.frozen = False
 
     @property
-    def runner(self):
-        return self._runner
+    def default_start_state(self):
+        return self._default_start_state
 
-    @property
-    def frozen(self):
-        return self._frozen
+    @default_start_state.setter
+    def default_start_state(self, state):
+        """Sets the *default* start state that the machine should use.
 
-    @property
-    def start_state(self):
-        return self._start_state
+        NOTE(harlowja): this will be used by ``initialize`` but only if that
+        function is not given its own ``start_state`` that overrides this
+        default.
+        """
+        if self.frozen:
+            raise excp.FrozenMachine()
+        if state not in self._states:
+            raise excp.NotFound("Can not set the default start state to"
+                                " undefined state '%s'" % (state))
+        self._default_start_state = state
 
     @property
     def current_state(self):
@@ -186,12 +194,12 @@ class FiniteMachine(object):
                                                 self._states[end]['on_enter'],
                                                 self._states[start]['on_exit'])
 
-    def process_event(self, event):
-        """Trigger a state change in response to the provided event."""
+    def _pre_process_event(self, event):
         current = self._current
         if current is None:
-            raise excp.NotInitialized("Can only process events after"
-                                      " being initialized (not before)")
+            raise excp.NotInitialized("Can not process event '%s'; the state"
+                                      " machine hasn't been initialized"
+                                      % event)
         if self._states[current.name]['terminal']:
             raise excp.InvalidState("Can not transition from terminal"
                                     " state '%s' on event '%s'"
@@ -200,23 +208,39 @@ class FiniteMachine(object):
             raise excp.NotFound("Can not transition from state '%s' on"
                                 " event '%s' (no defined transition)"
                                 % (current.name, event))
+
+    def _post_process_event(self, event, result):
+        return result
+
+    def process_event(self, event):
+        """Trigger a state change in response to the provided event."""
+        self._pre_process_event(event)
+        current = self._current
         replacement = self._transitions[current.name][event]
         if current.on_exit is not None:
             current.on_exit(current.name, event)
         if replacement.on_enter is not None:
             replacement.on_enter(replacement.name, event)
         self._current = replacement
-        return self._effect_builder(self._states[replacement.name], event)
+        result = self._effect_builder(self._states[replacement.name], event)
+        return self._post_process_event(event, result)
 
-    def initialize(self):
+    def initialize(self, start_state=None):
         """Sets up the state machine (sets current state to start state...)."""
-        if self._start_state not in self._states:
+        if start_state is None:
+            start_state = self._default_start_state
+        if start_state not in self._states:
             raise excp.NotFound("Can not start from a undefined"
-                                " state '%s'" % (self._start_state))
-        if self._states[self._start_state]['terminal']:
+                                " state '%s'" % (start_state))
+        if self._states[start_state]['terminal']:
             raise excp.InvalidState("Can not start from a terminal"
-                                    " state '%s'" % (self._start_state))
-        self._current = _Jump(self._start_state, None, None)
+                                    " state '%s'" % (start_state))
+        # No on enter will be called, since we are priming the state machine
+        # and have not really transitioned from anything to get here, we will
+        # though allow on_exit to be called on the event that causes this
+        # to be moved from...
+        self._current = _Jump(start_state, None,
+                              self._states[start_state]['on_exit'])
 
     def copy(self, shallow=False, unfreeze=False):
         """Copies the current state machine.
@@ -230,11 +254,12 @@ class FiniteMachine(object):
                         and want to use copies to run with (the copies have
                         the current state that is different between machines).
         """
-        c = FiniteMachine(self.start_state)
-        if unfreeze and self._frozen:
-            c._frozen = False
+        c = type(self)()
+        c._default_start_state = self._default_start_state
+        if unfreeze and self.frozen:
+            c.frozen = False
         else:
-            c._frozen = self._frozen
+            c.frozen = self.frozen
         if not shallow:
             for state, data in six.iteritems(self._states):
                 copied_data = data.copy()
@@ -253,7 +278,7 @@ class FiniteMachine(object):
 
     def freeze(self):
         """Freezes & stops addition of states, transitions, reactions..."""
-        self._frozen = True
+        self.frozen = True
 
     @property
     def states(self):
@@ -288,7 +313,7 @@ class FiniteMachine(object):
             if self.current_state == state:
                 prefix_markings.append("@")
             postfix_markings = []
-            if self.start_state == state:
+            if self.default_start_state == state:
                 postfix_markings.append("^")
             if self._states[state]['terminal']:
                 postfix_markings.append("$")
@@ -320,50 +345,6 @@ class FiniteMachine(object):
         return tbl.get_string()
 
 
-class _FiniteRunner(object):
-    """Finite machine runner used to run a finite machine."""
-
-    def __init__(self, machine):
-        self._machine = weakref.proxy(machine)
-
-    def run(self, event, initialize=True):
-        """Runs the state machine, using reactions only."""
-        for transition in self.run_iter(event, initialize=initialize):
-            pass
-
-    def run_iter(self, event, initialize=True):
-        """Returns a iterator/generator that will run the state machine.
-
-        NOTE(harlowja): only one runner iterator/generator should be active for
-        a machine, if this is not observed then it is possible for
-        initialization and other local state to be corrupted and cause issues
-        when running...
-        """
-        if initialize:
-            self._machine.initialize()
-        while True:
-            old_state = self._machine.current_state
-            reaction, terminal = self._machine.process_event(event)
-            new_state = self._machine.current_state
-            try:
-                sent_event = yield (old_state, new_state)
-            except GeneratorExit:
-                break
-            if terminal:
-                break
-            if reaction is None and sent_event is None:
-                raise excp.NotFound("Unable to progress since no reaction (or"
-                                    " sent event) has been made available in"
-                                    " new state '%s' (moved to from state '%s'"
-                                    " in response to event '%s')"
-                                    % (new_state, old_state, event))
-            elif sent_event is not None:
-                event = sent_event
-            else:
-                cb, args, kwargs = reaction
-                event = cb(old_state, new_state, event, *args, **kwargs)
-
-
 class HierarchicalFiniteMachine(FiniteMachine):
     """A fsm that understands how to run in a hierarchical mode."""
 
@@ -371,9 +352,10 @@ class HierarchicalFiniteMachine(FiniteMachine):
     _Effect = collections.namedtuple('_Effect',
                                      'reaction,terminal,machine')
 
-    @staticmethod
-    def _generate_runner(instance):
-        return _HierarchicalRunner(instance)
+    def __init__(self, default_start_state=None):
+        super(HierarchicalFiniteMachine, self).__init__(
+            default_start_state=default_start_state)
+        self._nested_machines = {}
 
     @classmethod
     def _effect_builder(cls, new_state, event):
@@ -389,97 +371,24 @@ class HierarchicalFiniteMachine(FiniteMachine):
             state, terminal=terminal, on_enter=on_enter, on_exit=on_exit)
         if machine is not None:
             self._states[state]['machine'] = machine
+            self._nested_machines[state] = machine
 
-    def initialize(self):
-        super(HierarchicalFiniteMachine, self).initialize()
+    def copy(self, shallow=False, unfreeze=False):
+        c = super(HierarchicalFiniteMachine, self).copy(shallow=shallow,
+                                                        unfreeze=unfreeze)
+        if shallow:
+            c._nested_machines = self._nested_machines
+        else:
+            c._nested_machines = self._nested_machines.copy()
+        return c
+
+    def initialize(self, start_state=None):
+        super(HierarchicalFiniteMachine, self).initialize(
+            start_state=start_state)
         for data in six.itervalues(self._states):
             if 'machine' in data:
                 data['machine'].initialize()
 
-
-class _HierarchicalRunner(object):
-    """Hierarchical machine runner used to run a hierarchical machine."""
-
-    def __init__(self, machine):
-        self._machine = weakref.proxy(machine)
-
-    def run(self, event, initialize=True):
-        """Runs the state machine, using reactions only."""
-        for transition in self.run_iter(event, initialize=initialize):
-            pass
-
-    @staticmethod
-    def _process_event(machines, event):
-        """Matches a event to the machine hierarchy.
-
-        If the lowest level machine does not handle the event, then the
-        parent machine is referred to and so on, until there is only one
-        machine left which *must* handle the event.
-
-        The machine whose ``process_event`` does not throw invalid state or
-        not found exceptions is expected to be the machine that should
-        continue handling events...
-        """
-        while True:
-            machine = machines[-1]
-            try:
-                result = machine.process_event(event)
-            except (excp.InvalidState, excp.NotFound):
-                if len(machines) == 1:
-                    raise
-                else:
-                    current = machine._current
-                    if current is not None and current.on_exit is not None:
-                        current.on_exit(current.name, event)
-                    machine._current = None
-                    machines.pop()
-            else:
-                return result
-
-    def run_iter(self, event, initialize=True):
-        """Returns a iterator/generator that will run the state machine.
-
-        This will keep a stack (hierarchy) of machines active and jumps through
-        them as needed (depending on which machine handles which event) during
-        the running lifecycle.
-
-        NOTE(harlowja): only one runner iterator/generator should be active for
-        a machine hierarchy, if this is not observed then it is possible for
-        initialization and other local state to be corrupted and causes issues
-        when running...
-        """
-        machines = [self._machine]
-        if initialize:
-            machines[-1].initialize()
-        while True:
-            old_state = machines[-1].current_state
-            effect = self._process_event(machines, event)
-            new_state = machines[-1].current_state
-            try:
-                machine = effect.machine
-            except AttributeError:
-                pass
-            else:
-                if machine is not None and machine is not machines[-1]:
-                    machine.initialize()
-                    machines.append(machine)
-            try:
-                sent_event = yield (old_state, new_state)
-            except GeneratorExit:
-                break
-            if len(machines) == 1 and effect.terminal:
-                # Only allow the top level machine to actually terminate the
-                # execution, the rest of the nested machines must not handle
-                # events if they wish to have the root machine terminate...
-                break
-            if effect.reaction is None and sent_event is None:
-                raise excp.NotFound("Unable to progress since no reaction (or"
-                                    " sent event) has been made available in"
-                                    " new state '%s' (moved to from state '%s'"
-                                    " in response to event '%s')"
-                                    % (new_state, old_state, event))
-            elif sent_event is not None:
-                event = sent_event
-            else:
-                cb, args, kwargs = effect.reaction
-                event = cb(old_state, new_state, event, *args, **kwargs)
+    @property
+    def nested_machines(self):
+        return self._nested_machines
